/**************************************************************/
/************* NYSTROM ****************************************/
/**************************************************************/
/*    Irene Rodriguez-Lujan    *********************************/
/*    Ramon Huerta			  *********************************/
/*    Charles Elkan			  *********************************/
/*    Carlos Santa Cruz		  *********************************/
/**************************************************************/
/** June 3rd, 2010                                        ****/
/** Based on "Quadratic Programming Feature Selection"       **/
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "NYSTROM.h"



#define EPSILON	1e-6



/* Funciones LAPACK */


int build_Ais(long k,double *ais,double *u, double *S);
int  build_Q(long d, long k, double *q, double *a, double *asi, double *b) ;
void mult(long m, long l, long n,double *c, double *a, double *b);
void mult_trans(long m,long l, long n,double *c, double *a, double *b);



/* private functions */
short NYSTROM_CalcularEigen(NYSTROM_pNYSTROM x);


/******************************************************************/
/****** funcion de diagonalizacion por el metodo Nystrom      ***/
NYSTROM_pNYSTROM NYSTROM_CalcularNYSTROM(double **K, long d, long k)

{

  NYSTROM_pNYSTROM  x=NULL;
  long i,j;



  srand(6713589);


  /* reserva memoria para el nystrom */
 
  if (!(x=NYSTROM_AlocarNYSTROM(d,k))) {
    fprintf(stderr,"ERROR: memory allocation of Nystrom Structure\n");
    return NULL;
  }


  /* Asigna matriz */
  for (i=0;i<k;i++)
	for (j=0;j<d;j++)
  		x->MCova[i][j]=K[i][j];



 
  /* diagonaliza la matriz segun el metodo Nystrom */

  if (NYSTROM_CalcularEigen(x)==NYSTROM_ERROR) {
    NYSTROM_LiberarNYSTROM(&x);
    return NULL;
  }


  /* devuelve el resultado */
  return x;

}


/**********************************************************************/
/**	      diagonalizacion de la matriz de covarianza          *********/
/**    salida 0: error de alocacion			                      *********/
/**	      1: salida correcta			                            *********/
/**	      2: no converge				                              *********/
/**********************************************************************/

short NYSTROM_CalcularEigen(NYSTROM_pNYSTROM x)

{

  long		i,j,r;
  double * A = NULL, *B=NULL, *V=NULL, *Ais=NULL, *Q=NULL, *C=NULL, *D=NULL, *E=NULL, *S=NULL;
  


  /* Aplicacion del metodo Nystrom */

  /* Se forma la matriz A (se supone de rango maximo) */
  
  
  A = (double *) malloc ((x->k)*(x->k)*sizeof(double));


  if (!A)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix A (%ld)\n",(x->k)*(x->k));
	return NYSTROM_ERROR;
  }

  B = (double *) malloc (x->k*((x->d)-(x->k))*sizeof(double));

   
  if (!B)
  {
	fprintf(stderr,"ERROR: memory allocation ofo matrix B (%ld)\n",x->k*((x->d)-(x->k)));
	return NYSTROM_ERROR;
  }

  

  for (i=0;i<(x->k);i++)
  {
	for (j=0;j<(x->k);j++)
	{
		A[i*(x->k)+j]=x->MCova[i][j];
	}
	

	for (j=(x->k);j<(x->d);j++)
	{		
		B[i*(x->d-x->k)+(j-x->k)]=x->MCova[i][j];	
	}
	
  }

	
  V=(double *) malloc (x->k*x->k*sizeof(double));

  if (!V)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix V (%ld)\n",x->k*x->k);
	return NYSTROM_ERROR;
  }


  S=(double *) malloc(x->k*sizeof(double));
  
  if (!S)
  {
	fprintf(stderr,"ERROR: memory allocation of vector S (%ld)\n",x->k);
	return NYSTROM_ERROR;
  }


  // r: contiene el rango de la matriz 
  r=svd(x->k,A,V,S);



  /* Se reconstruye A porque hara falta mas adelante */
  for (i=0;i<(x->k);i++)
  {
	for (j=0;j<(x->k);j++)
	{
		A[i*(x->k)+j]=x->MCova[i][j];
	}
  }


  Ais = (double *) malloc (x->k*x->k*sizeof(double));


  if (!Ais)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'Ais' (%ld) \n",x->k*x->k);
	return NYSTROM_ERROR;
  }

 
  i = build_Ais(x->k,Ais,V,S);

  
  Q = (double *) malloc (x->k*x->k*sizeof(double));

  if (!Q)
  {
	fprintf(stderr,"ERROR al reservar memoria para la matriz Q\n");
	return NYSTROM_ERROR;
  }


  i = build_Q(x->d,x->k,Q,A,Ais,B);


  i=svd(x->k,Q,V,S); 


  /* Una vez que los autovalores estan en w, hay que calcular
	los autovectores
  */


  C = (double *) malloc (x->d*x->k*sizeof(double));

  if (!C)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix C (%ld)\n",x->d*x->k);
	return NYSTROM_ERROR;
  }


  for (i=0;i<(x->k);i++)
  {
	for (j=0;j<(x->d);j++)
	{			
		C[j*(x->k)+i]=x->MCova[i][j];		
	}	
  }


  /* Esta multiplicacion sencilla no merece la pena hacerla con mult */

  D = (double *) calloc (x->k*x->k,sizeof(double));

  if (!D)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix D (%ld)\n",x->k*x->k);
	return NYSTROM_ERROR;
  }
  


  for (i=0;i<(x->k);i++)
  {
	for (j=0;j<(x->k);j++)
		//IRENE: tolerancia como en Matlab	tol = max (size (X)) * sigma_max (X) * eps,		
		if (S[j]>(S[0]*EPSILON))
			//gsl_matrix_set(D,i,j,gsl_matrix_get(V,i,j)*(1.0/sqrt(gsl_vector_get(S,j))));
			D[i*(x->k)+j]= V[j*(x->k)+i]*(1.0/sqrt(S[j]));
  }

  mult(x->k,x->k,x->k,Q,Ais,D);


  E = (double *) calloc (x->d*x->k,sizeof(double));
 
  if (!E)
  {
	fprintf(stderr,"ERROR al reservar la matriz E\n");
	return NYSTROM_ERROR;
  } 


  mult(x->d,x->k,x->k,E,C,Q);
 

  /* Se mueve el resultado a la estructura */

  for(i=0;i<(x->k);i++)
	x->eval[i]=S[i];

  for (i=0;i<(x->d);i++)
  {

	for (j=0;j<(x->k);j++)		
		x->evec[i][j]=E[i*(x->k)+j];

  }


  /* Liberar  memoria */
 
  free (A);
  free (B);
  free (C);
  free (D);
  free (E);
  free (V);
  free(Q);
  free(S);
  free(Ais);
  



  return NYSTROM_OK;
  
}




/********************************************************
*				AUXILIARY FUNCTIONS						*
********************************************************/

/***
* C CON LAPACK
***/
int svd(long k,double *a,double *u,double *s) {
 
  char opc[1];
  int m,n,ldu,ldv,lwork,lda;
  double *v=NULL;
  int info=0;
  double *work=NULL;
  int *iwork=NULL;


  
  /*** TRANSFORMACIONES PARA FORTRAN ***/

  lda=ldu=ldv=m=n=k;
  opc[0]='A';

  v=(double *) malloc (n*n*sizeof(double));

  if(!v)
  {
	fprintf(stderr,"ERROR: memory allocation of matrix 'v' in svd\n");
	return -1;
  }

  lwork=11*n*n;

  work=(double *)malloc(lwork*sizeof(double));

  if(!work)
  {
	fprintf(stderr,"ERROR: memory allocation of vector 'work' in svd\n");
	return -1;
  }


  iwork = (int *)malloc(8*n*sizeof(int));

  if(!iwork)
  {
	fprintf(stderr,"ERROR: memory allocation of vector 'iwork' in svd\n");
	return -1;
  }


  dgesdd_(opc,&m,&n,a,&lda,s,u,&ldu,v,&ldv,work,&lwork,iwork,&info);




  free(v);
  free(work);
  free(iwork);

  return info;
}




/***
* Irene: r es el rango de la matriz 
**/

int build_Ais(long k, double *ais,double *u, double *S){
  
  int l,m;
  double *c=NULL;  


  c = (double *) calloc (k*k,sizeof(double));

  if (c==NULL) {
	fprintf(stderr,"ERROR: memory allocation of matrix 'c'\n");
	return NYSTROM_ERROR;
  }

  
  for(l=0;l<k;++l) {   
    for(m =0;m<k;++m) {	
	if (S[m]>EPSILON*S[0])
	{	
		
		c[l*k+m]=u[m*k+l]*(1.0/sqrt(S[m]));		
	}
    }
  }

  mult(k,k,k,ais,c,u);


  free(c);

  return NYSTROM_OK;

}


int  build_Q(long d,long k,double *q, double *a, double *asi, double *b)
{


	double *tmp=NULL, *tmp2=NULL;
	long i,j;



	tmp = (double *) malloc (k*k*sizeof(double));

	if (!tmp)
	{
		fprintf(stderr,"ERROR: memory allocation of matrix 'tmp'\n");
		return NYSTROM_ERROR;
	}

	tmp2 = (double *) malloc (k*k*sizeof(double));

	if (!tmp2)
	{
		fprintf(stderr,"ERROR: memory allocation of matrix 'tmp2'\n");
		return NYSTROM_ERROR;
	}

	
	mult_trans(k,d-k,k,tmp,b,b);

	mult(k,k,k,tmp2,asi,tmp);

	mult(k,k,k,q,tmp2,asi);


	/* Q+A */
	for (i=0;i<k;i++)
		for (j=0;j<k;j++)
			q[i*k+j]+=a[i*k+j];
	
	free(tmp);
	free(tmp2);


	return NYSTROM_OK;
}

void mult(long m, long l, long n,double *c, double *a, double *b) {
  int i,j,k;
  double sum;

  for(i=0;i<m;++i) {
    for(j=0;j<n;++j) {
      sum=0.0;
      for(k=0;k<l;++k) 
	//sum+=gsl_matrix_get(a,i,k)*gsl_matrix_get(b,k,j);
	sum+=a[i*l+k]*b[k*l+j];
      //gsl_matrix_set(c,i,j,sum);
      c[i*n+j]=sum;
    }
  }
}

/** 
* a => mxl
* b => nxl
**/

void mult_trans(long m, long l, long n,double *c, double *a, double *b) {
  int i,j,k;
  double sum;

  
  for(i=0;i<m;++i) {
    for(j=0;j<n;++j) {      
      sum=0.0;
      for(k=0;k<l;++k) {
		//sum+=gsl_matrix_get(a,i,k)*gsl_matrix_get(b,j,k);
		sum+=a[i*l+k]*b[j*l+k];
	}
      //gsl_matrix_set(c,i,j,sum);
      c[i*n+j]=sum;
    }
  }
 
}


/************************************************************
***********		INPUT/OUTPUT FUNCTIONS	*********************
*************************************************************/
/***********************************************/
/* Alocacion de la estructura NYSTROM              */
/***********************************************/


NYSTROM_pNYSTROM NYSTROM_AlocarNYSTROM(long d, long k)

{
  
  NYSTROM_pNYSTROM	x=NULL;
  long i;


  /* reserva la estructura */
  if (!(x=(NYSTROM_pNYSTROM)calloc(1,sizeof(NYSTROM_nystrom)))) {
    fprintf(stderr,"ERROR: allocation of NYSTROM structure\n");
    return NULL;
  }

  /* asigna las variables de la estructura y reserva memoria */
  x->d=d;
  x->k=k;
  x->autoCorrelacion=0;
  


  x->MCova = (double **) malloc (x->d*sizeof(double));
  if (!x->MCova)
  {
    NYSTROM_LiberarNYSTROM(&x);
    fprintf(stderr,"ERROR: memory allocation of matrix 'MCova'\n");
    return NULL;
  }

  for (i=0;i<(x->d);i++)
  {
	x->MCova[i]=(double *)malloc(x->d*sizeof(double));
	if (!x->MCova[i])
	{
 		NYSTROM_LiberarNYSTROM(&x);
    	return NULL;
	}
  }


  x->evec=(double **) malloc (d*sizeof(double*));
  if (!x->evec)
  {
	fprintf(stderr,"ERROR: memory allocation in NYSTROM\n");
	NYSTROM_LiberarNYSTROM(&x);
    return NULL;
  }
  
  for (i=0;i<d;i++)
  {
	x->evec[i]=(double *)malloc (k*sizeof(double));
	if(!x->evec[i])
	{
	  fprintf(stderr,"ERROR: memory allocation in NYSTROM\n");
	  NYSTROM_LiberarNYSTROM(&x);
	  return NULL;
	}
  }

  x->eval=(double *) malloc (k*sizeof(double));
  if (!x->eval)
  {
	fprintf(stderr,"ERROR: memory allocation in NYSTROM\n");
	NYSTROM_LiberarNYSTROM(&x);
    return NULL;
  }

 
 x->media=(double *) malloc (d*sizeof(double));
 if (!x->media)
 {
   fprintf(stderr,"ERROR: memory allocation in NYSTROM\n");
   NYSTROM_LiberarNYSTROM(&x);
   return NULL;   
 }
 
  x->varianza=(double *) malloc (d*sizeof(double));
  if (!x->varianza)
  {
	fprintf(stderr,"ERROR: memory allocation in NYSTROM\n");
	NYSTROM_LiberarNYSTROM(&x);
    return NULL;	
  }


  /* salida correcta */
  return x;

}



/***********************************************/
/* Liberar la estructura de datos de NYSTROM       */
/***********************************************/

void NYSTROM_LiberarNYSTROM(NYSTROM_pNYSTROM *x)

{
  long i;

  if (x && (*x)) {


   if ((*x)->MCova)
   {
	for (i=0;i<((*x)->d);i++)
	{
		if ((*x)->MCova[i])
		{
			free((*x)->MCova[i]);
			(*x)->MCova[i]=NULL;
		}
	}
	free((*x)->MCova);
	(*x)->MCova=NULL;
   }

    if ((*x)->evec) {
      for (i=0;i<(*x)->d;i++)
	  {
		if ((*x)->evec[i])
		{
		  free((*x)->evec[i]);
		  (*x)->evec[i]=NULL;
		}
	  }
	  free((*x)->evec);
      (*x)->evec=NULL;
    }
	
    if ((*x)->eval) {
      free((*x)->eval);
      (*x)->eval=NULL;
    }
    if ((*x)->media) {
      free((*x)->media);
      (*x)->media=NULL; 
    }
    if ((*x)->varianza) {
      free((*x)->varianza);
      (*x)->varianza=NULL; 
    }
    free(*x);
    (*x)=NULL;
  }

}



