/**************************************************************/
/************* NYSTROM.c ****************************************/
/**************************************************************/
/*    Irene Rodriguez-Lujan    *********************************/
/*    Ramon Huerta			  *********************************/
/*    Charles Elkan			  *********************************/
/*    Carlos Santa Cruz		  *********************************/
/**************************************************************/
/** June 3rd, 2010                                        ****/
/** Based on "Quadratic Programming Feature Selection"       **/
/**************************************************************/

#ifndef NYSTROM_H
#define NYSTROM_H


#define NYSTROM_ERROR	0
#define NYSTROM_OK		1
#define NYSTROM_LINE_LENGTH 100000
#define NYSTROM_TQLI_ITER_MAX 300
#define NYSTROM_TINY  1E-16


typedef struct _NYSTROM_nystrom {
  long		d;		              /* dimension de los auto... */
  short   autoCorrelacion;    /* 0/1 matriz covarizanzas/autocorrelacion */
  double	**MCova;	          /* matriz de covarianzas */
  double	**evec; 	          /* puntero a la matriz de autovectores */
  double	*eval;		          /* punetro a los autovalores */
  double	*media; 	          /* puntero a las medias */
  double  *varianza;          /* puntero a las varianzas */
  long    k;			/* Dimension de la submatriz Nystrom */
} NYSTROM_nystrom, *NYSTROM_pNYSTROM;

/*
 * Liberacion y alocacion de la memoria de la estructura de datos
 */


void NYSTROM_LiberarNYSTROM(NYSTROM_pNYSTROM *x);
NYSTROM_pNYSTROM NYSTROM_AlocarNYSTROM(long d, long k);

// extern void  dgesvd_( char *, char *, int *, int *, double *, int *, double *, double *
//   		       , int *, double *,int *,double *, int *, int *);
extern void  dgesdd_(char *, int *, int *, double *, int *, double *, double *
  		       , int *, double *,int *,double *, int *, int *,int *);
// extern void sgesvd_( char *, char *, int *, int *, double *, int *, double *, double *
//  		       , int *, double *,int *,double *, int *, int *);

int svd(long k,double *a,double *v2,double *s);


/*
 * Metodo Nystrom
 */

NYSTROM_pNYSTROM NYSTROM_CalcularNYSTROM(double **K, long d, long k);


#endif

