/**************************************************************/
/************* QP.h ****************************************/
/**************************************************************/
/*    Irene Rodriguez-Lujan    *********************************/
/*    Ramon Huerta			  *********************************/
/*    Charles Elkan			  *********************************/
/*    Carlos Santa Cruz		  *********************************/
/**************************************************************/
/** June 3rd, 2010                                        ****/
/** Based on "Quadratic Programming Feature Selection"       **/
/**************************************************************/

#ifndef QP_QUADRATICPROG_H
#define QP_QUADRATICPROG_H



#define QP_ERROR	0
#define QP_OK		1
#define QP_LINE_LENGTH 10000000
#define QP_TINY  1E-16
#define QP_TQLI_ITER_MAX 500
#define QP_VAR_TINY	1E-10
#define QP_MAX_VARIANZA 1E+50
#define EPSILON	1e-6


typedef struct _QP_qp {
  long		d;		/* dimension */
  double	alpha;	/* parameter relevance-redundancy */
  double	**Q;	/* quadratic term matrix */
  double	*F;		/* linear term vector */
  double	*h;		/* solution vector */
  double 	*eval;	/* eigenvalues */
  double 	**evec;	/* eigenvectors */
} QP_qp, *QP_pQP;

/*
 * Allocation QP structure
 */


void QP_FreeQP(QP_pQP *x);
QP_pQP QP_AllocQP(long d,double alpha);


/*
 * Quadratic Programming
 */

short QP_EntropyMatrix (QP_pQP qp,double **data, long nData, long d, long nClases, long nSegment, long nSigma, long * aleat, long ns,double alpha);
short QP_CorrelationMatrix (QP_pQP qp,double **data, long nData, long d, long nClasses, long * aleat, long ns, double alpha);
short QP_SolveQP(QP_pQP qp, long ns);
void QP_SortQP(QP_pQP qp, long *permutacion);





/*
 * funciones I/O
 */

short QP_WriteQP(QP_pQP qp, char *FicheroSal, long submuestreos, long *aleat,long *sort,long offset);

#endif

